//==============================================================================
//  Component:
//
//  Module:
//      $KW; $FN; $EKW;
//      $KW; $Own; $Ver; $ChkD; $EKW;
//
//  Purpose:
//
// [Keywords:]
//
//  Copyright (C) 1999-2001 Interstron, Ltd.
//  All rights reserved.
//==###=========================================================================

#include <setjmp.h>
#include <stdlib.h>
#include <exception>
#include "TryCatch.h"

#ifdef __cplusplus
    extern "C" {
#endif


    struct ObjectTable* _curObjectTable = NULL;
    struct TryBlock* _curTryBlock = NULL;
    void* _curThrownObject = NULL;
    ThrownObject* _curThrownObjectInfo = NULL;

    static char buf[256];
    static unsigned bufUsed;

    static void* _allocate(unsigned size)
    {
        if(sizeof(buf) - bufUsed < size)
        {
            void* ptr = malloc(size);
            if(!ptr) std::terminate();
            return ptr;
        }
        else
        {
            void* ptr = buf + bufUsed;
            bufUsed += size;
            bufUsed = (bufUsed + (sizeof(int) - 1)) & ~(sizeof(int) - 1);
            return ptr;
        }
    }

    static void _deallocate(void* ptr)
    {
        if(ptr >= buf && ptr < buf + bufUsed) bufUsed = (char*)ptr - buf;
        else free(ptr);
    }

    void* _allocateThrownObject(unsigned size)
    {
        struct ThrownObject* except = (struct ThrownObject*)_allocate(sizeof(struct ThrownObject) + size);

        except->obj.nObjects = 1;
        except->obj.next = NULL;
        except->obj.objects[0]._this = except + 1;
        except->obj.objects[0].dtor = NULL;
        except->nextTry = _curTryBlock;
        _curTryBlock = NULL;
        except->type = NULL;
        except->next = _curThrownObjectInfo;
        _curThrownObjectInfo = except;
        except->rethrow = 0;

        return except->obj.objects[0]._this;
    }

    static int _canCatch(struct TypeInfo* thrown, struct TypeInfo* catched)
    {
        for(int i = 0;; i++)
        {
            if(thrown == catched) return 1;

            if((thrown->tag & etPointer) && (catched->tag & etPointer))
            {
                thrown = thrown->next;
                catched = catched->next;

                if((catched->tag & etConst) < (thrown->tag & etConst) ||
                    (catched->tag & etVolatile) < (thrown->tag & etVolatile))
                    return 0;

                if(((catched->tag & etTerm) || (catched->tag & etClass)) &&
                    ((catched->tag & etConst) || (catched->tag & etVolatile)))
                    catched = catched->next;

                if(((thrown->tag & etTerm) || (thrown->tag & etClass)) &&
                    ((thrown->tag & etConst) || (thrown->tag & etVolatile)))
                    thrown = thrown->next;

                if(thrown == catched || catched == &VOID_CATCH_INFO) return 1;

                continue;
            }

            if(i <= 1 && (catched->tag & etClass) && (thrown->tag & etClass))
            {
                struct TypeInfoClass* ptr = (struct TypeInfoClass*)thrown;
                for(int i = 0; i < ptr->size; i++)
                    if((struct TypeInfo*)ptr->bases[i].type == catched)
                    {
                        if(i == 0)
                        {
                            if(ptr->bases[i].vbase != -1)
                            {
                                _curThrownObject = *(void**)((char*)_curThrownObject + ptr->bases[i].vbase);
                            }
                            _curThrownObject = (char*)_curThrownObject + ptr->bases[i].offset;
                        }
                        else
                        {
                            if(*(void**)_curThrownObject != NULL)
                            {
                                if(ptr->bases[i].vbase != -1)
                                {
                                    _curThrownObject = *(void**)(*(char**)_curThrownObject + ptr->bases[i].vbase);
                                }
                                *(void**)_curThrownObject = *(char**)_curThrownObject + ptr->bases[i].offset;
                            }
                        }
                        return 1;
                    }
            }

            break;
        }

        return 0;
    }

    void _throwObject(struct TypeInfo* thrown, void (*dtor)(void*))
    {
        int i;
        struct TryBlock* temp;

        if(thrown)
        {
            _curThrownObjectInfo->obj.objects[0].dtor = dtor;
            _curThrownObjectInfo->type = thrown;
            temp = _curThrownObjectInfo->nextTry;
        }
        else if(_curThrownObjectInfo)
        {
            _curThrownObjectInfo->rethrow++;
            thrown = _curThrownObjectInfo->type;
            temp = _curTryBlock;
            _curTryBlock = NULL;
        }
        else temp = NULL;

        while(temp)
        {
            while(_curObjectTable != temp->objs)
            {
                while(_curObjectTable->nObjects)
                {
                    if(_curThrownObjectInfo->obj.objects[0]._this == _curObjectTable->objects[i]._this)
                    {
                        _curThrownObjectInfo->rethrow--;
                        break;
                    }

                    i = --_curObjectTable->nObjects;

                    if(_curObjectTable->objects[i]._this && _curObjectTable->objects[i].dtor)
                        _curObjectTable->objects[i].dtor(_curObjectTable->objects[i]._this);
                }

                _curObjectTable = _curObjectTable->next;
            }

            _curThrownObject = _curThrownObjectInfo->obj.objects[0]._this;
            for(i = 0; i < temp->catches->size; i++)
            {
                if(!temp->catches->types[i] || _canCatch(thrown, temp->catches->types[i]))
                {
                    _curThrownObjectInfo->nextTry = temp->next;
                    longjmp(temp->jbuf, i + 1);
                }
            }

            temp = temp->next;
        }

        std::terminate();
    }

    void _exceptionCatched()
    {
        _curTryBlock = _curThrownObjectInfo->nextTry;
        _curThrownObjectInfo->nextTry = NULL;

        if(!_curThrownObjectInfo->rethrow)
        {
            _curThrownObjectInfo->obj.next = _curObjectTable;
            _curObjectTable = &_curThrownObjectInfo->obj;
        }
    }

    void _destroyThrownObject()
    {
        if(!_curThrownObjectInfo->rethrow)
        {
            struct ThrownObject* temp = _curThrownObjectInfo;
            _curThrownObjectInfo = temp->next;

            if(temp->obj.objects[0].dtor) temp->obj.objects[0].dtor(temp->obj.objects[0]._this);
            _deallocate(temp);

            while(_curThrownObjectInfo && !_curThrownObjectInfo->obj.nObjects)
            {
                temp = _curThrownObjectInfo;
                _curThrownObjectInfo = temp->next;
                _deallocate(temp);
            }
        }
        else
        {
            _curThrownObjectInfo->rethrow--;
        }
    }

#ifdef __cplusplus
    }
#endif

//==###=========================================================================
//  End of module:
//      $KW; $FN; $EKW;
//==============================================================================
