//////////////////////////////////////////////////////////////////////////
//                                                                      //
// powf.cpp -                                                           //
//                                                                      //
//                              //
//        .                          //
//              pow(x, y) = exp( y * log(x) )
//                                                                      //
// Copyright (c) 2001 RC Module Inc.                                    //
//                                                                      //
//  :  ..                                                 //
//                                                                      //
//   $Revision:: 1     $      $Date:: 2/14/06 8:09p    $                //
//                                                                      //
//////////////////////////////////////////////////////////////////////////

#include "math.h"
#include "auxfd.h"

//   
//  Ln(1+1/2)*c, Ln(1+1/4)*c, Ln(1+1/8)*c, ...
//  c = 040000000
static unsigned int _TLN[29]={0x19F323EC,	0xE47FBE3,	0x789C1DB,	0x3E14618,
		0x1F829B0,	0xFE0545,	0x7F80A9,	0x3FE015,
		0x1FF802,	0xFFE00,	0x7FF80,	0x3FFE0,
		0x1FFF8,	0xFFFE,	0x7FFF,	0x3FFF,	0x1FFF,	0xFFF,	0x7FF,
		0x3FF,	0x1FF,	0xFF,	0x7F,	0x3F,	0x1F,	0xF,	0x7,
		0x3,	0x1	};

//    
//	    ln(x) = ln(~=2)+ln(2^(-1)) = ln(~=2)-ln(2)
static unsigned int _TMLN[28]={	0x12696211,	0x11178e82,	0x108598b5,	0x10415d89,	0x10205658,
				0x10101575,	0x10080559,	0x10040155,	0x10020055,	0x10010015,
				0x10008005,	0x10004001,	0x10002000,	0x10001000,	0x10000800,
				0x10000400,	0x10000200,	0x10000100,	0x10000080,	0x10000040,
				0x10000020,	0x10000010,	0x10000008,	0x10000004,	0x10000002,
				0x10000001,	0x10000000,	0x10000000	};

//  Ln(2) * 0x40000000
const unsigned int _TLN2 = 0x2C5C85FE;

//       ,   [0,1]
inline unsigned int _MulFix0x64(unsigned int M1, unsigned int M2, int& NewExp){
	unsigned int rez=0;
	while (!(M1 & 0x80000000)){
		NewExp--;
		M1<<=1;
	}
	while (!(M2 & 0x80000000)){
		NewExp--;
		M2<<=1;
	}
	while( M2>>=1 ){
        if (M1 & 0x80000000)
            rez += M2;
        M1<<=1;
	}
    if (rez & 0x80000000){
        rez>>=1;
		NewExp++;
    }
	return rez;
}

//    
//   -  CORDIC
inline unsigned int SrvLog(float f, int& NewExp, int& OldSgn, int& Sgn){

	int Mnt;                           //   
	int Exp, Spc;                      //   
	unsigned int  Arg;                 //     

    //   float
	DestructFloat( f, Mnt, Exp, OldSgn, Spc);

	//  
	if ( Exp ==0x80 )	// -x, +inf, NAN:  NAN
		return 0x40000001;

	if ( f == 0 )	// 0 -> +inf
		return 0x40000000;

	Mnt <<= 7;
	unsigned int tmp; //       
	int rez = 0;
	Arg= 0x40000000; /*   1,  
								        */
    Sgn = 0;
	NewExp = 0;
	int i=0,j=0;

	tmp = Arg;

	if (Exp==-1){
		Mnt>>=1;
		while (i<27){  
			i++;
			tmp -= (tmp>>(i+1)) + ((tmp>>(i))&1); 
			if (tmp>=Mnt){
				if (j==0){
					j=i;
					NewExp=-j+1;
					rez<<=j;
				}
				Arg = tmp;
				rez += _TMLN[i-1]>>(i-j);

				tmp -= (tmp>>(i+1)) + ((tmp>>(i))&1); 
				if (tmp>=Mnt){
					Arg = tmp;
					rez += _TMLN[i-1]>>(i-j);
				}
				else
					tmp = Arg;
			}
			else
				tmp = Arg;

		}
		Exp=0;
		Sgn=1;
	}
	else
		while (i<27){  
			i++;
			tmp += tmp>>i; 
			if (tmp<=Mnt){
				Arg = tmp;
				rez += _TLN[i-1];
			}
			else 
				tmp = Arg;
		}
	
//	int tln2 = _TLN2;

	if (Exp<-1){
		rez = -rez;
		Exp = -Exp;
		Sgn=1;
	}

	i=0;
	int NewMnt = rez;
	while (Exp>0){  
		i++;
		if (Exp & 1){
			NewMnt += _TLN2;
			rez = NewMnt;
			rez >>=1;
			NewExp = i;
		}
		Exp>>=1;
		NewMnt>>=1;
	}
    if (rez & 0x40000000){
		NewExp++;
		rez >>=1;
    }

	return rez;
}

//   
//   -  CORDIC
inline float Exp(unsigned int Mnt, int Exp, int Sgn){

	int NewExp;
	unsigned int Arg = Mnt;

	int i=-Exp;
    if (i>0){
        Arg>>=i;
        i=0;
    }
	//     () ,  floor 

    int tmp; //       
	unsigned int rez = 0x40000000; /*   1,  
								        */
	NewExp = 0;
	tmp = Arg;

	while (i<0){  //  
		i++;
		tmp -= _TLN2; 
		if (tmp>0){
			Arg = tmp;
			NewExp += 1 << (-i);
		}
		else //    - 
			tmp = Arg;
		tmp <<=1;
		Arg = tmp;
	}
	tmp >>=1;
	Arg = tmp;
	if (NewExp>0x7f){
		if (Sgn==1){
			if(NewExp>0x7f){
				if (!_FPIEEE_Flags)
					_FPIEEE_Signal( _FPIEEE_Underflow ); 
				return ConstructFloat( 0, -0x7f, 0, 0);
			}
		}
		else{
			if (!_FPIEEE_Flags)
				_FPIEEE_Signal( _FPIEEE_Overflow ); 
			return ConstructFloat( 0, 0x80, 0, 0);	
		}
	}
	if (Sgn==1){
		Arg = _TLN2 - Arg;
		tmp = Arg;
		NewExp = -(++NewExp);
	}
//	}
	while (i!=62){  //  
		tmp -= _TLN[i++]; //    ,
//    ,  _TLN[]
//  ,    
		if (tmp>0){
			rez += rez >> i; //  - !
			Arg = tmp;
		}
		else //    - 
			tmp = Arg;
	}

	if (NewExp<-0x7e){	//  -  
		rez >>= (-0x7f - NewExp)+1;
		NewExp = -0x7f;
	}

	rez >>=6;	//    , ...
	rez = (rez>>1) + (rez & 1); //   ,  .

	if (rez == 0x01000000){	
		rez = 0;
		NewExp++;
	}

	return ConstructFloat( rez, NewExp, 0, 0);	
}

#ifdef __cplusplus
extern "C" {
#endif

// powf --
//
//           
//
//      --.
//      CORDIC,    .
float powf (float X, float Y){   // pow(x, y) = exp( y * ln(x) )
    int XExp, XSgn, LogSgn;
    int YMnt;
	int YExp, YSgn;
	int NewExp, Spc;

    //  ,   
    //  float Dbg_Res = Log(X); int Sgn;
    //	DestructFloat( Dbg_Res, YMnt, YExp, YSgn, Spc);

    unsigned int XMnt = SrvLog(X, XExp, XSgn, LogSgn );    // ln(x)
	if ( XMnt ==0 )
		return 1;

	unsigned int Mul;
	DestructFloat( Y, YMnt, YExp, YSgn, Spc);

	//  
	if ( YExp ==0 && YMnt==0x00800000 && YSgn ==0 )	// y = 1
		return X;

	if ( YExp == -0x7f && YMnt==0 )	// y = 0
		return ConstructFloat( 0, 0, 0, 0);	

    if ( XExp ==0x80 )	
        if (XMnt == 0){             // X = inf
		    if (YSgn == 0)
                return X;
            else                    
                return ConstructFloat( 0, -0x7f, XSgn, 0);
        }
        else{                       // X = NAN
            return X;
        }

    if ( XSgn ){    //   
        if ( YExp < 0 )    // NAN
    		return ConstructDouble( 1, 0x80, 0, 0);	
        else if ( YExp > 23){
            XSgn = 0;
        }
        else if ( (( YMnt>>(23-YExp) )<<(23-YExp)) != YMnt ){   // Y - , NAN
    		return ConstructFloat( 1, 0x400, 0, 0);	
        }
        else if ( (( YMnt>>(24-YExp) )<<(24-YExp)) == YMnt )   // Y - , |X|^Y
            XSgn = 0;
                                                               //      -(|X|^Y)
    }
    NewExp = XExp + YExp;
    Mul =  _MulFix0x64(XMnt, YMnt, NewExp)>>1;
    NewExp += 11;

    //       
    //  Dbg_Res = log(X) * Y;
    //	DestructFloat( Dbg_Res, YMnt, XExp, Sgn, Spc);
    //   double R1=ConstructFloat( YMnt, XExp, Sgn, 0);
    //     return exp (ConstructFloat( Mul>>6, NewExp+1, YSgn^LogSgn, 0));

    if (XSgn)
        return -Exp(Mul, NewExp, YSgn^LogSgn );
    else
        return Exp(Mul, NewExp, YSgn^LogSgn );
}

#ifdef __cplusplus
}
#endif
